import path from "path";
import fsp from "fs/promises";
import fs from "fs";
import toc from "hyperui-docs/toc";
import matter from "gray-matter";

/**
 * Function that reads all the files in /src/docs/ and generates a complete README.md
 */
async function main() {
    let readme = "<!-- This README is auto-generated. Please do not modify this file. --> \n\n";

    // Generate table of contents
    readme += "# Table of Contents \n";
    readme += await renderTableOfContents(toc, 1);
    readme += "\n";

    // Turn table of contents into flat array
    let tocFlat = tocToFlatArray(toc);

    // Iterate through the table of contents and load all the files
    for (const entry of tocFlat) {
        readme += await renderPage(entry);
    }

    // Write the README.md file to the root directory
    await fsp.writeFile(path.join(process.cwd(), "README.md"), readme);
}

main();

async function pathExists(_path) {
    try {
        await fsp.access(_path, fs.constants.F_OK);
        return true;
    } catch (error) {
        return false;
    }
}

async function renderTableOfContents(currentDict, level = 1) {
    let tocRendered = "";

    for (const key in currentDict) {
        // Stay at level
        if (typeof currentDict[key] == "string") {
            let filePath = path.join(process.cwd(), "../docs", currentDict[key]);
            let fileText = await fsp.readFile(filePath, "utf-8");
            let { data: frontmatter } = matter(fileText);

            let newLink = `#${frontmatter.title?.toLowerCase().replace(/ /g, "-")}`;

            // Build indentation
            let identation = "";
            for (let i = 0; i < level; i++) {
                identation += "  ";
            }

            // Append the toc entry to the list
            tocRendered += `${identation} - [${key}](${newLink}) \n`;
        }

        // Traverse deeper
        if (typeof currentDict[key] == "object") {
            tocRendered += await renderTableOfContents(currentDict[key], level + 1);
        }
    }

    return tocRendered;
}

async function renderPage(entry) {
    let filePath = path.join(process.cwd(), "../docs", entry.slug);
    let fileText = await fsp.readFile(filePath, "utf-8");
    let { content: mdContent, data: frontmatter } = matter(fileText);

    // Replace relative links with markdown title-based links
    let links = mdContent.match(/\[(.*?)\]\((.*?)\)/g) ?? [];
    for (const link of links) {
        let linkName = link.match(/\[(.*?)\]/)[1];
        let targetPathRel = link.match(/\((.*?)\)/)[1];
        let targetPathAbs = path.join(path.dirname(filePath), targetPathRel);

        if ((await pathExists(targetPathAbs)) == false) {
            console.log("Skipping link: ", targetPathRel);
            continue;
        }

        let targetText = await fsp.readFile(targetPathAbs, "utf-8");
        let { data: targetFrontmatter } = matter(targetText);

        let newLink = `#${targetFrontmatter.title?.toLowerCase().replace(/ /g, "-")}`;
        mdContent = mdContent.replace(link, `[${linkName}](${newLink})`);
    }

    // Append the content to the README file
    return `# ${frontmatter.title} \n${mdContent} \n`;
}

function tocToFlatArray(current) {
    let result = [];

    for (const key in current) {
        if (Object.prototype.hasOwnProperty.call(current, key)) {
            const value = current[key];

            if (typeof value == "object" && value !== null) {
                // Recursively traverse nested objects
                result.push(...tocToFlatArray(value));
            } else {
                // Push non-object values to the result array
                result.push({
                    title: key,
                    slug: value,
                });
            }
        }
    }

    return result;
}
