import Color from "color";

/**
 * Notes on Variants:
 * There is alyways a primary, secondary, ghost and ghost-bg variant.
 * For all provided colors, there will be a solid and ghost/ghost-bg variant computed.
 */

export function compileBrandTheme(theme) {
    // Deep copy
    theme = JSON.parse(JSON.stringify(theme));

    // Validate theme
    if (theme.colors === undefined) throw new Error("HyperUI: Missing colors in brand theme.");
    if (Object.keys(theme.colors).length === 0) throw new Error("HyperUI: No colors defined in brand theme.");
    if (theme.accentColor === undefined) throw new Error("HyperUI: Missing accent color in brand theme.");
    if (theme.colors[theme.accentColor] === undefined)
        throw new Error(`HyperUI: Accent color ${theme.accentColor} not found in brand theme.`);

    // Compile variants (-> compute missing values and apply real color values)
    Object.entries(theme.colors).forEach(([colorName, colorValue]) => {
        let _colorValue = Color(colorValue);

        let _ghost_fontColor = "";
        let _ghost_backgroundColor = "";
        let _ghost_hoverColor = "";
        let _ghost_borderColor = "";

        let _solid_fontColor = "";
        let _solid_backgroundColor = "";
        let _solid_hoverColor = "";
        let _solid_borderColor = "transparent";
        let _solid_borderColorPressed = "transparent";

        // Validate color
        if (colorValue === undefined || !_colorValue) throw new Error(`HyperUI: Invalid color value for ${colorName}: ${colorValue}`);

        // Compute colors for ghost variants
        _ghost_fontColor = colorValue;
        _ghost_backgroundColor = calculateGhostBackgroundColorOnFontColor(_ghost_fontColor);
        _ghost_hoverColor = calculateHoverColorOnBackgroundColor(_ghost_backgroundColor);
        _ghost_borderColor = colorValue;

        // Compute colors for solid variants
        _solid_backgroundColor = colorValue;
        _solid_hoverColor = calculateHoverColorOnBackgroundColor(_solid_backgroundColor);

        if (_colorValue.isDark() || _colorValue.luminosity() < 0.5) _solid_fontColor = "var(--hy-font-color-white)";
        else _solid_fontColor = "var(--hy-font-color-black)";

        if (_colorValue.luminosity() > 0.8) {
            _solid_borderColor = "var(--hy-element-border-secondary-color)";
            _solid_borderColorPressed = "var(--hy-accent-color)";
        }

        // Apply colors to variables
        theme.colors[colorName] = {
            color: colorValue,
            ghost: {
                fontColor: _ghost_fontColor,
                backgroundColor: _ghost_backgroundColor,
                hoverColor: _ghost_hoverColor,
                borderColor: _ghost_borderColor,
            },
            solid: {
                fontColor: _solid_fontColor,
                backgroundColor: _solid_backgroundColor,
                hoverColor: _solid_hoverColor,
                borderColor: _solid_borderColor,
                borderColorPressed: _solid_borderColorPressed,
            },
        };
    });

    // Copy accent color to "primary" variant and trim ghost variant
    theme.colors["primary"] = { ...theme.colors[theme.accentColor], ghost: null };

    // Compile accent color
    theme.accentColor = theme.colors[theme.accentColor].color;

    // Akkumulate all available element variants
    theme.availableVariants = [];

    Object.entries(theme.colors).forEach(([colorName, colorConfig]) => {
        if (colorConfig.solid) theme.availableVariants.push(colorName);
        if (colorConfig.ghost) {
            theme.availableVariants.push(`ghost-${colorName}`);
            theme.availableVariants.push(`ghost-bg-${colorName}`);
        }
    });

    return theme;
}

export function generateCssVariables(theme) {
    let variables = {};

    // Generate CSS variables for variants
    Object.entries(theme.colors).forEach(([colorName, colorConfig]) => {
        // Ghost variant
        if (colorConfig.ghost) {
            variables[`--hy-variant-ghost-${colorName}-font-color`] = colorConfig.ghost.fontColor;
            variables[`--hy-variant-ghost-${colorName}-bg-color`] = colorConfig.ghost.backgroundColor;
            variables[`--hy-variant-ghost-${colorName}-hover-color`] = colorConfig.ghost.hoverColor;
            variables[`--hy-variant-ghost-${colorName}-border-color`] = colorConfig.ghost.borderColor;
        }

        // Solid variant
        if (colorConfig.solid) {
            variables[`--hy-variant-${colorName}-font-color`] = colorConfig.solid.fontColor;
            variables[`--hy-variant-${colorName}-bg-color`] = colorConfig.solid.backgroundColor;
            variables[`--hy-variant-${colorName}-hover-color`] = colorConfig.solid.hoverColor;
            variables[`--hy-variant-${colorName}-border-color`] = colorConfig.solid.borderColor;
            variables[`--hy-variant-${colorName}-border-color-pressed`] = colorConfig.solid.borderColorPressed;
        }
    });

    // Generate CSS variables for colors
    Object.entries(theme.colors).forEach(([colorName, colorConfig]) => {
        if (colorConfig) {
            variables[`--hy-color-${colorName}`] = colorConfig.color;
        }
    });

    // Set accent color
    variables["--hy-accent-color"] = theme.accentColor;
    variables["--hy-accent-hover-color"] = calculateHoverColorOnBackgroundColor(theme.accentColor);

    return variables;
}

export function generateCssUtilityClasses(theme) {
    let utilityClasses = "";

    Object.entries(theme.colors).forEach(([colorName]) => {
        utilityClasses += `

            .hy-text-${colorName} {
                &,
                * {
                    color: var(--hy-color-${colorName}) !important;
                }
            }

            .hy-bg-${colorName} {
                background-color: var(--hy-color-${colorName});
            }
        `;
    });

    return utilityClasses;
}

export function applyCssVariables(variables) {
    Object.entries(variables).forEach(([key, value]) => {
        document.documentElement.style.setProperty(key, value);
    });
}

export function applyCssUtilityClasses(utilityClasses) {
    let styleElement = document.createElement("style");
    styleElement.id = "hy-utility-classes-inject";
    styleElement.innerHTML = utilityClasses;
    document.body.appendChild(styleElement);
}

// Helper functions
function calculateHoverColorOnBackgroundColor(backgroundColor) {
    backgroundColor = Color(backgroundColor);

    if (backgroundColor.isDark() || backgroundColor.luminosity() < 0.5) {
        backgroundColor = backgroundColor.darken(0.15);
    } else {
        backgroundColor = backgroundColor.darken(0.05);
    }

    backgroundColor = backgroundColor.opaquer(1.2);
    return backgroundColor.hexa();
}

function calculateGhostBackgroundColorOnFontColor(fontColor) {
    fontColor = Color(fontColor);
    fontColor = fontColor.fade(0.9);
    return fontColor.hexa();
}
